/*
 * Copyright (c) 1980, 1988 The Regents of the University of California.
 * All rights reserved.  [see below]
 *
 * "MAIN LOOP" section:  Copyright (c) 1994-2004 Greg Roelofs
 *
 * GRR revision history:
 *    2 May 1994  created
 *   25 Jul 1996  [unknown]
 *   31 Dec 1996  fix up "No mail" message; define NO_PATHS_H for Sun
 *    5 Jun 1997  sendmail 8.* compatibility
 *   21 Nov 1997  added Content-Header capability to avoid segfaults
 *    2 Jan 1998  added ToLower/zstrnicmp() for "content-header"
 *   25 May 1998  worked around buggy "From" addresses, e.g., Crocker
 *   20 Jun 1998  added support for MAIL variable [envfile]
 *   18 Jan 2000  added -d option to suppress date for longer subjects
 *   15 Feb 2000  fixed -f option so not overridden by MAIL variable
 *   27 Jan 2002  added DEBUG; added header var. (avoid bogus content-lengths)
 *    3 May 2002  large modifications to support line- and byte-counts
 *    7 May 2002  rewrote main loop; fixed line/byte-counting logic
 *   16 Jun 2003  added subject-sanitization (e.g., spam with embedded CRs)
 *   18 Jun 2003  added -c option for colorized output
 *    7 Feb 2004  added support for "-f -" (stdin as input stream)
 *   20 Mar 2004  added from-sanitization (e.g., spam with embedded CRs)
 *   27 Jun 2005  added hh:mm -> yyyy overwrite if not current year
 *   22 Jan 2006  added support for wide windows (> 80 columns)
 *
 * GRR TO DO:  check isatty() on input and use to avoid "-f -" requirement
 * GRR TO DO:  colorize "known good" senders (e.g., svlug, etc.); spam score
 * GRR TO DO:  telescope runs of white space to single space while stuffing buf
 *
 * To compile under SunOS:  cc -O -o from_grr from_grr.c -s
 * To compile under IRIX:   cc -cckr -O -o from_grr from_grr.c -s
 * To compile under Linux:  gcc -Wall -O3 -o from_grr from_grr.c
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1980, 1988 The Regents of the University of California.\n\
 Portions copyright (c) 1994-2006 Greg Roelofs.  All rights reserved.\n";
/* static */ char sccsid[] =
"@(#)from_grr.c    5.6 (Berkeley/Newtware) 2006-01-22";
#endif /* not lint */

#include <sys/types.h>
#include <ctype.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>		/* GRR 19980620:  for getenv() */
#include <unistd.h>		/* GRR 20020503:  for getopt() */
#include <string.h>		/* GRR 199xxxxx:  for strlen() in main loop */
#include <time.h>		/* GRR 20050627:  for time(), localtime() */
#include <sys/ioctl.h>		/* GRR 20060122:  for ioctl() (screenwidth) */

#define CMDLINE_OPTION_WINS	/* GRR 20000215:  -f overrides MAIL variable */

#ifdef DEBUG			/* GRR 20020507:  usually fprintf... */
#  define Trace(x)   printf x
#else
#  define Trace(x)
#endif

#ifndef TRUE
#  define TRUE   1
#  define FALSE  0
#endif

#ifdef sun
#  define NO_PATHS_H
#endif

#ifdef NO_PATHS_H
#  ifndef _PATH_MAILDIR
#    define _PATH_MAILDIR  "/usr/spool/mail"
#  endif
#else
#  include <paths.h>
#endif
/* GRR 19960725: should make this dependent on actual _PATH_MAILDIR, but lazy */
#define _PATH_ALTMAILDIR   "/usr/mail"

#if MAXPATHLEN > BUFSIZ
#  define BUFSZ  MAXPATHLEN
#else
#  define BUFSZ  BUFSIZ
#endif
/* GRR 20020502:  not actually checking for long lines, so make buffer BIG: */
#if BUFSZ < 16384
#  undef BUFSZ
#  define BUFSZ  16384
#endif
/* GRR 20020503:  short buffers for fields that get printed out: */
#define SBUFSZ   512

/* GRR 20060122:  for wide windows, limit "from" part; rest goes to subject */
#define MAX_FROMLEN_NORMAL    35
#define MAX_FROMLEN_LONG  42

#define ToLower(x)   ((char)(isupper((int)x)? tolower((int)x) : x))

/* GRR 20030618:  new -c option to colorize parts of output */
#define COLOR_NORMAL        "\033[0m"
#define COLOR_RED_BOLD      "\033[01;31m"	/* [spam score 3/10 -pobox] */
#define COLOR_RED           "\033[40;31m"
#define COLOR_GREEN_BOLD    "\033[01;32m"
#define COLOR_GREEN         "\033[40;32m"
#define COLOR_YELLOW_BOLD   "\033[01;33m"	/* sender (frombuf) */
#define COLOR_YELLOW        "\033[40;33m"	/* recipient (To:) -- D.N.E. */
#define COLOR_BLUE_BOLD     "\033[01;34m"
#define COLOR_BLUE          "\033[40;34m"
#define COLOR_MAGENTA_BOLD  "\033[01;35m"
#define COLOR_MAGENTA       "\033[40;35m"
#define COLOR_CYAN_BOLD     "\033[01;36m"
#define COLOR_CYAN          "\033[40;36m"
#define COLOR_WHITE_BOLD    "\033[01;37m"	/* subject (subjbuf) */
#define COLOR_WHITE         "\033[40;37m"	/* date (datebuf) */

int match();
int zstrnicmp();		/* GRR 19980102 */
unsigned getscreenwidth();	/* GRR 20060122 */


int main(argc, argv)
    int argc;
    char **argv;
{
    extern char *optarg;
    extern int optind;
    struct passwd *pwd, *getpwuid();
    int ch, emptyline, color=FALSE, skipdate=FALSE, maxsubjlen=31;
    int header;					/* GRR 20020127 */
    int len, linenum, last_linenum, msgnum;	/* GRR 20020127, 20020503 */
    int have_fromc, have_subject;		/* GRR 20020507 */
    int linelen=0, have_content_length;		/* GRR 20020507 */
    long content_length, content_left=0L;	/* GRR 20020507 */
    long bytenum, last_bytenum;			/* GRR 20020503 */
    unsigned datelen, subjlen, fromlen;		/* GRR 20060122 */
    unsigned screenwidth;			/* GRR 20060122 */
    char *file, *barefile="null", *envfile, *sender, *p;
    char buf[BUFSZ];
    char frombuf[SBUFSZ], datebuf[SBUFSZ], subjbuf[SBUFSZ], outbuf[SBUFSZ*2];

    file = sender = NULL;
    while ((ch = getopt(argc, argv, "cdf:s:")) != EOF) {
        switch((char)ch) {
        case 'c':		/* GRR 20030618 */
            color = TRUE;
            break;
        case 'd':		/* GRR 20000118 */
            skipdate = TRUE;
            maxsubjlen = 49;
            break;
        case 'f':
            file = optarg;
            break;
        case 's':
            sender = optarg;
            for (p = sender; *p; ++p)
                if (isupper(*p))
                    *p = tolower(*p);
            break;
        case '?':
        default:
            fprintf(stderr,
              "usage: from_grr [-cd] [-f file] [-s sender] [user]\n");
            fprintf(stderr,
              "  -c  colorize output (sender, date, subject, etc.)\n");
            fprintf(stderr,
              "  -d  omit date from output in favor of more of subject line\n");
            fflush(stderr);
            return 1;
        }
    }
    argv += optind;

    if (!file) {
        if (!(file = *argv)) {
            if (!(pwd = getpwuid(getuid()))) {
                fprintf(stderr,
                  "from_grr: no filename and no password file entry for you.\n"
                  );
                fflush(stderr);
                return 1;
            }
            barefile = pwd->pw_name;
        }
        (void)sprintf(buf, "%s/%s", _PATH_MAILDIR, barefile);
        file = buf;
    }
    if (strcmp(file, "-") != 0) {	/* GRR 20040207:  already have stdin */
#ifdef CMDLINE_OPTION_WINS		/* GRR 20000215 */
        if (file != buf) {
            if (!freopen(file, "r", stdin)) {
                fprintf(stderr, "from_grr: can't read %s.\n", file);
                fflush(stderr);
                return 1;
            }
        } else
#endif /* CMDLINE_OPTION_WINS */
        if ((envfile = getenv("MAIL")) != NULL && !freopen(envfile, "r", stdin))
        {
            if (!freopen(file, "r", stdin)) {
#ifndef CMDLINE_OPTION_WINS	/* GRR 20000215 */
                if (file != buf) {
                    fprintf(stderr, "from_grr: can't read %s.\n", file);
                    fflush(stderr);
                    return 1;
                }
#endif /* !CMDLINE_OPTION_WINS */
                /* _PATH_MAILDIR/name didn't work; try _PATH_ALTMAILDIR/name */
                (void)sprintf(buf, "%s/%s", _PATH_ALTMAILDIR, barefile);
                if (!freopen(file, "r", stdin)) {
                    if (envfile)
                        fprintf(stderr,
                          "No mail (can't read %s, %s/%s or %s).\n",
                          envfile, _PATH_MAILDIR, barefile, file);
                    else
                        fprintf(stderr, "No mail (can't read %s/%s or %s).\n",
                          _PATH_MAILDIR, barefile, file);
                    fflush(stderr);
                    return 1;
                }
            }
        }
    }

    last_linenum = 1;
    linenum = msgnum = 0;
    bytenum = last_bytenum = 0L;
    emptyline = TRUE;   /* empty lines delimit messages and header blocks */
    header = FALSE;     /* not until see "^From " with emptyline && !header */
    have_content_length = FALSE;
    have_fromc = FALSE;
    have_subject = FALSE;

    /* GRR 20060122 */
    fromlen = skipdate? 25 : 21;
    datelen = 16;
    subjlen = skipdate? 48 : 23;
    screenwidth = getscreenwidth();
    if (screenwidth > 80) {
        unsigned max_fromlen = skipdate? MAX_FROMLEN_LONG : MAX_FROMLEN_NORMAL;
        unsigned extra = screenwidth - 80;
        unsigned from_extra = extra * fromlen/(fromlen + subjlen);

        fromlen += from_extra;
        if (fromlen > max_fromlen) {
            from_extra -= (fromlen - max_fromlen);
            fromlen = max_fromlen;
        }
        subjlen += (extra - from_extra);
    }

/*---------------------------------------------------------------------------
    START OF MAIN LOOP (over lines)
  ---------------------------------------------------------------------------*/

    while (fgets(buf, sizeof(buf), stdin)) {
        ++linenum;
        linelen = strlen(buf);   /* save for content-length handling, output */
        bytenum += linelen;

                        /*===================================================*/
        if (header) {   /*                 >>>>> HEADER <<<<<                */
                        /*===================================================*/

            /* emptyline indicates a completely empty line (regex = "^$"),
             *  which is the delimiter for end-of-message and end-of-header
             *  (as used, variable means "the last line was empty") */
            if (*buf == '\n') {
                emptyline = TRUE;
                header = FALSE;      /* end of header, by definition */
                Trace(("GRR DEBUG:  line %d = empty (end of header)\n",
                  linenum));
                continue;
            } else                   /* variable not used within header, */
                emptyline = FALSE;   /*  so safe to update it here */

            Trace(("GRR DEBUG:  line %d:  within header\n", linenum));

            if (!have_content_length &&
                zstrnicmp(buf, "Content-Length: ", 16) == 0)
            {
                have_content_length = TRUE;
                sscanf(buf+16, "%ld", &content_length);
                Trace(("GRR DEBUG:  content-length = %ld bytes:  will skip\n",
                  content_length));
                content_left = content_length;
                continue;
            }

            if (!have_fromc && strncmp(buf, "From: ", 6) == 0) {
                char *pFrom = buf + 6;

                have_fromc = TRUE;
                len = strlen(pFrom);
                if (len > fromlen)
                    len = fromlen;          /* _max_ displayed length... */
                pFrom[len] = '\0';          /* truncate at 25 characters */
                if (pFrom[len-1] == '\n')
                    pFrom[len-1] = '\0';    /* lose the newline */
                strcpy(frombuf, pFrom);
                /* GRR 20040320:  sanitize From line of CRs, etc. (also spam) */
                for (pFrom = frombuf;  *pFrom;  ++pFrom) {
                    if (*pFrom < ' ' || *pFrom > '~')
                        *pFrom = '|';
                }
                continue;
            }

            if (!have_subject && strncmp(buf, "Subject: ", 9) == 0) {
                char *pSubj = buf + 9;

                have_subject = TRUE;
                len = strlen(pSubj);
                if (len > subjlen)
                    len = subjlen;               /* _max_ displayed length... */
                pSubj[len] = '\0';
                if (pSubj[len-1] == '\n')
                    pSubj[len-1] = '\0';    /* lose the newline */
                strcpy(subjbuf, pSubj);
                /* GRR 20030616:  sanitize subject line of CRs, etc. (spam) */
                for (pSubj = subjbuf;  *pSubj;  ++pSubj) {
                    if (*pSubj < ' ' || *pSubj > '~')
                        *pSubj = 'X';
                }
                continue;
            }

            /* (don't care about any other header types) */

                   /*========================================================*/
        } else {   /*                     >>>>> BODY <<<<<                   */
                   /*========================================================*/

            /* skip content to avoid bogus "From" and "Content-Length" lines */
            /* GRR 20020502:  was fseek(stdin, content_length, SEEK_CUR); */

            if (have_content_length) {
                content_left -= linelen;
                if (content_left <= 0L) {
                    content_left = content_length = 0L;
                    have_content_length = FALSE;
                }
                if (*buf == '\n') {
                    emptyline = TRUE;   /* "last line was empty" */
                    Trace((
                      "GRR DEBUG:  line %d = empty (within content-length)\n",
                      linenum));
                } else {
                    emptyline = FALSE;
                    Trace(("GRR DEBUG:  line %d:  within content-length\n",
                      linenum));
                }
                continue;
            }

            if (*buf == '\n') {
                emptyline = TRUE;   /* "last line was empty" */
                Trace(("GRR DEBUG:  line %d = empty (within body)\n", linenum));
                continue;
            }

            /* "From " header == start of new message if previous line empty */

            if (emptyline && strncmp(buf, "From ", 5) == 0 &&
                (!sender || match(buf + 5, sender)))  /* -s option (untested) */
            {
                char *p, *pDate, *pFrom=buf+5;

                /* GRR 20020503:  OK, we just entered a new header, so...first
                 * calculate length of previous message (lines, bytes), if any,
                 * and print out requested type of output; then reset length
                 * variables for next message; and finally, do normal sender
                 * and date processing on "From " line itself */

                /* XXX bad assumption:  old sprintf()s return POINTER: */
                if (msgnum > 0) {
                    if (skipdate) {
                        len = sprintf(outbuf, "%3d %-*s  ", msgnum, fromlen, frombuf);
                        /* this is crude, but what the heck:  need proper len */
                        if (color) {
                            /* overwrite uncolored version; don't change len */
                            sprintf(outbuf, "%3d %s%-*s%s  ", msgnum,
                              COLOR_YELLOW, fromlen, frombuf, COLOR_NORMAL);
                        }
                    } else {
                        /* "from" part is shorter (21) to accommodate lengths */
                        frombuf[fromlen] = '\0';
                        /* don't count current line (linelen) in byte length */
                        len = sprintf(outbuf, "%3d %-*s  %s %4d/%-6ld ",
                          msgnum, fromlen, frombuf, datebuf, linenum - last_linenum,
                          bytenum - last_bytenum - linelen);
                        if (color) {
                            /* overwrite uncolored version; don't change len */
                            sprintf(outbuf, "%3d %s%-*s%s  %s%s%s "
                              "%s%4d%s/%s%-6ld%s ", msgnum,
                              COLOR_YELLOW, fromlen, frombuf, COLOR_NORMAL,
                              COLOR_MAGENTA, datebuf, COLOR_NORMAL,
                              COLOR_CYAN_BOLD, linenum - last_linenum,
                              COLOR_NORMAL, /* / */ COLOR_CYAN,
                              bytenum - last_bytenum - linelen, COLOR_NORMAL);
                        }
                    }
                    subjbuf[screenwidth-1-len] = '\0';   /* avoid too-long lines: 79 max */
                    if (color) {
                        printf("%s%s%s%s\n", outbuf, COLOR_WHITE, subjbuf,
                          COLOR_NORMAL);
                    } else {
                        printf("%s%s\n", outbuf, subjbuf);
                    }
                    fflush(stdout);   /* GRR 19971121:  sometimes segfault... */

                    last_linenum = linenum;
                    last_bytenum = bytenum - linelen;
                }

                ++msgnum;
                header = TRUE;
                emptyline = FALSE;
                have_content_length = FALSE;
                have_fromc = FALSE;
                have_subject = FALSE;

                subjbuf[0] = '\0';               /* in case there isn't one */
                if ((pDate = strchr(pFrom, ' ')) == NULL) {
                    strcpy(datebuf, "[ unknown date ]");
                    strncpy(frombuf, buf+5, fromlen); /* save in case no "From " */
                    frombuf[fromlen] = '\0';
                } else {
                    *pDate++ = '\0';             /* truncate pFrom, just i/c */
                    strncpy(frombuf, pFrom, fromlen); /* ...and copy, just in case */
                    frombuf[fromlen] = '\0';
                    if (*pDate == ' ')
                        ++pDate;
                    p = strrchr(pDate, ':');     /* one colon always in there */
                    if (p[-3] != ':')            /* ...but sometimes only one */
                        p += 3;
                    *p = '\0';  /* lose secs (if any), year/timezone, newline */
                    /* GRR 19980525: */
                    len = strlen(pDate);
                    if (len > datelen) {  /* pretty sure last 16 are OK; lose head */
                        pDate += (len - datelen);
                        len = datelen;
                    }
#if 0

    |     21 chars      |  |   16 chars   |             |      23 chars       |
262 Greg Roelofs <newt@po  Fri Jan 20 18:37   61/2969   Re: Four tuned paramete

    |       25 chars        |  |                   48 chars                   |
262 Greg Roelofs <newt@pobox.  Re: Four tuned parameters xxxxxxxxxxxxxxxxxxxxxx

12345678901234567890123456789012345678901234567890123456789012345678901234567890
        10        20        30        40        50        60        70        80

                    if (pDate[len-1] == '\n')
                        pDate[len-1] = '\0';   /* lose the newline */
#endif
                    /* GRR 20050627:  if year != current, use that, not hh:mm */
                    if ((p = strchr(p+1, ' ')) != NULL) {
                        if (!isdigit(p[1])) {
                            /* maybe "PDT" or something?  try one more time */
                            p = strchr(p+1, ' ');
                        }
                        if (p && isdigit(p[1])) {
                            int yr = atoi(p+1);
                            if (yr > 1900 && yr < 2200) {
                                time_t now = time(NULL);
                                int yr_cur = localtime(&now)->tm_year + 1900;
                                if (yr != yr_cur) {
                                    /* overwrite "hh:mm" with " yyyy" */
                                    strncpy(pDate+len-5, p, 5);
                                }
                            }
                        }
                    }
                    strncpy(datebuf, pDate, datelen);
                    datebuf[datelen] = '\0';
                }
                continue;

            } /* end if ("From " header) */

            /* (don't care about anything else in body */

            emptyline = FALSE;  /* already looped if TRUE; now safe to update */
        }
    }

/*---------------------------------------------------------------------------
    END OF MAIN LOOP (over lines)
  ---------------------------------------------------------------------------*/

    /* GRR 20020503:  print final message (if any) */
    if (msgnum > 0) {
        /* XXX bad assumption:  old sprintf()s (e.g., DEC) return POINTER: */
        if (skipdate) {
            len = sprintf(outbuf, "%3d %-*s  ", msgnum, fromlen, frombuf);
            /* this is crude, but what the heck:  need proper len */
            if (color) {
                /* overwrite uncolored version, but don't change len */
                sprintf(outbuf, "%3d %s%-*s%s  ", msgnum,
                  COLOR_YELLOW, fromlen, frombuf, COLOR_NORMAL);
            }
        } else {
            /* make "from" slightly shorter (21) to accommodate lengths */
            frombuf[fromlen] = '\0';
            /* add 1 to line count because we're not in 1st line of next msg */
            /* (also, no need to "correct" for linelen since don't have one) */
            len = sprintf(outbuf, "%3d %-*s  %s %4d/%-6ld ", msgnum,
              fromlen, frombuf, datebuf, linenum - last_linenum + 1,
              bytenum - last_bytenum);
            if (color) {
                /* overwrite uncolored version; don't change len */
                sprintf(outbuf, "%3d %s%-*s%s  %s%s%s %s%4d%s/%s%-6ld%s ",
                  msgnum, COLOR_YELLOW, fromlen, frombuf, COLOR_NORMAL,
                  COLOR_MAGENTA, datebuf, COLOR_NORMAL,
                  COLOR_CYAN_BOLD, linenum - last_linenum + 1, COLOR_NORMAL,
                  /* / */
                  COLOR_CYAN, bytenum - last_bytenum, COLOR_NORMAL);
            }
        }
        subjbuf[screenwidth-1-len] = '\0';   /* avoid too-long lines: 79 max */
        if (color) {
            printf("%s%s%s%s\n", outbuf, COLOR_WHITE, subjbuf,
              COLOR_NORMAL);
        } else {
            printf("%s%s\n", outbuf, subjbuf);
        }
        fflush(stdout);   /* GRR 19971121:  sometimes segfault... */
    }

    return 0;
}



int match(line, sender)
    register char *line, *sender;
{
    register char ch, pch, first, *p, *t;

    for (first = *sender++;;) {
        if (isspace(ch = *line))
            return(0);
        ++line;
        if (isupper(ch))
            ch = tolower(ch);
        if (ch != first)
            continue;
        for (p = sender, t = line;;) {
            if (!(pch = *p++))
                return(1);
            if (isupper(ch = *t++))
                ch = tolower(ch);
            if (ch != pch)
                break;
        }
    }
    /* NOTREACHED */
}



/************************/
/* Function zstrnicmp() */  /* copyright Info-ZIP (see UnZip 5.3x, fileio.c) */
/************************/

int zstrnicmp(s1, s2, n)
    register char *s1, *s2;
    register int n;
{
    for (; n > 0;  --n, ++s1, ++s2) {

        if (ToLower(*s1) != ToLower(*s2))
            /* test includes early termination of one string */
            return (ToLower(*s1) < ToLower(*s2))? -1 : 1;

        if (*s1 == '\0')   /* both strings terminate early */
            return 0;
    }
    return 0;
}



/*****************************/
/* Function getscreenwidth() */ /* from BSD mailx 5.5 (height stuff removed) */
/*****************************/

unsigned getscreenwidth()
{
    unsigned screenwidth;
    struct winsize ws;

    if (ioctl(1, TIOCGWINSZ, (char *) &ws) < 0)
        ws.ws_col = ws.ws_row = 0;

    if ((screenwidth = ws.ws_col) == 0) {
        // GRR FIXME:  also check COLUMNS env variable
        //printf("GRR DEBUG:  ioctl(TIOCGWINSZ) failed\n"); fflush(stdout);
        screenwidth = 80;
    }

    return screenwidth;
}



/*
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
